/*
 * Copyright 1995 by the Massachusetts Institute of Technology.
 * All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice appear in all
 * copies and that both that copyright notice and this permission
 * notice appear in supporting documentation, and that the name of
 * M.I.T. not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior
 * permission.  M.I.T. makes no representations about the suitability
 * of this software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 */
/* %%%EOOH%%% */

#include <stdio.h>
#include <ctype.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#ifdef SOLARIS
#include <sys/filio.h>
#endif
#include <sys/param.h>
#include <netdb.h>
#include <errno.h>
#include "inet-udp.h"
#include <arpa/inet.h> /* hmmm... */
#include "Connect.h"

#ifndef NULL
#define NULL 0
#endif

extern char *malloc();

Trap udp_init(), udp_open(), udp_close(), udp_ntoa(), udp_aton(),
  udp_free(), udp_send(), udp_receive();

ConnectDomain inetudp =
{
  "inet-udp",
  udp_init,
  udp_ntoa,
  udp_aton,
  udp_free,
  udp_open,
  udp_close,
  udp_send,
  udp_receive
};

udpaddr *addrhash[INETTABLESIZE], *inethash[INETTABLESIZE];
int udpdomain;
Addr wild;
int nameResolve;

/* REGISTER: Connect_RegisterDomain(&inetudp); */

#define hash_addr(a) ((a)%INETTABLESIZE)
#define hash_inet(h) (((h)->sin_port + (h)->sin_addr.s_addr)%INETTABLESIZE)
#define inet_eq(a, b) (((a)->sin_port == (b)->sin_port) && \
		       ((a)->sin_addr.s_addr == (b)->sin_addr.s_addr))

Trap udp_init(d)
     int d;
{
  int i;

  udpdomain = d;
  for (i = 0; i < INETTABLESIZE; i++)
    {
      addrhash[i] = NULL;
      inethash[i] = NULL;
    }

  wild = Connect_AnyAddress(d);
  nameResolve = 1;
  return OK;
}

udpaddr *addrlookup(addr)
     Addr addr;
{
  udpaddr *a;

  a = addrhash[hash_addr(addr)];

  while (a != NULL)
    {
      if (addr == a->address)
	break;
      a = a->nextaddr;
    }

  return a;
}

udpaddr *inetlookup(inetaddr)
     struct sockaddr_in *inetaddr;
{
  udpaddr *a;

  a = inethash[hash_inet(inetaddr)];

  while (a != NULL)
    {
      if (inet_eq(inetaddr, &(a->netaddr)))
	break;
      a = a->nextinet;
    }

  return a;
}

Trap newaddress(new, addr)
     udpaddr **new, *addr;
{
  udpaddr *newaddr;
  int ahash, ihash;
  Addr a;

  newaddr = (udpaddr *)malloc(sizeof(udpaddr));
  if (newaddr == NULL)
    Return(GENERAL_MALLOC, sizeof(udpaddr));
  memcpy(newaddr, addr, sizeof(udpaddr));

  if (Connect_NewAddress(udpdomain, &a))
    {
      free(newaddr);
      return CHECK;
    }

  ahash = hash_addr(a);
  ihash = hash_inet(&(addr->netaddr));

  newaddr->address = a;
  newaddr->nextaddr = addrhash[ahash];
  newaddr->nextinet = inethash[ihash];
  addrhash[ahash] = newaddr;
  inethash[ihash] = newaddr;

  *new = newaddr;
  return OK;
}

void freeaddr(freeaddr)
     udpaddr *freeaddr;
{
  int ahash, ihash;
  udpaddr *find;

  ahash = hash_addr(freeaddr->address);
  ihash = hash_inet(&(freeaddr->netaddr));

  /*
   * Remove from address bucket...
   */
  if (addrhash[ahash] == freeaddr)
    addrhash[ahash] = freeaddr->nextaddr;
  else
    {
      find = addrhash[ahash];
      while (find != NULL && find->nextaddr != freeaddr)
	find = find->nextaddr;
      if (find->nextaddr == freeaddr)
	find->nextaddr = freeaddr->nextaddr;
    }

  /*
   * Remove from inet bucket...
   */
  if (inethash[ihash] == freeaddr)
    inethash[ihash] = freeaddr->nextinet;
  else
    {
      find = inethash[ihash];
      while (find != NULL && find->nextinet != freeaddr)
	find = find->nextinet;
      if (find->nextinet == freeaddr)
	find->nextinet = freeaddr->nextinet;
    }

  if (freeaddr->hostname != NULL)
    free(freeaddr->hostname);
  free(freeaddr);
}

/*
 * ERRORS:
 *
 * CONNECT_BADADDRESSFORMAT: FATAL
 * The address string %s is ill-formed.
 *
 * CONNECT_UNKNOWNNAME: FATAL
 * The hostname %s was not resolvable.
 *
 * CONNECT_DOMAINMISMATCH: FATAL
 * The hostname given did not match the domain it was paired with.
 *
 * CONNECT_NEWADDRESS: INFO
 * Operation generated a new address (%d).
 *
 * UDP_UNKNOWNSERVICE: FATAL
 * Service name requested for address (%s/udp) was unknown.
 */
Trap udp_ntoa(domain, name, address)
     int domain;
     char *name;
     Addr *address;
{
  struct hostent *hp;
  struct servent *sp;
  char *n, *p;
  udpaddr addr, *s;

  if (name[0] == '*' && name[1] == '\0')
    {
      *address = wild;
      return OK;
    }

  n = malloc(strlen(name) + 1);
  if (n == NULL)
    Return(GENERAL_MALLOC, strlen(name) + 1);
  strcpy(n, name);

  memset(&addr, 0, sizeof(addr));
  addr.netaddr.sin_family = AF_INET;

  p = strchr(n, '/');
  if (p == NULL)
    addr.netaddr.sin_port = 0;
  else
    {
      *p = '\0';
      p++;
      if (isdigit(*p))
	addr.netaddr.sin_port = htons(atoi(p));
      else
	{
	  sp = getservbyname(p, "udp");
	  if (sp == NULL)
	    {
	      memmove(n, p, strlen(p) + 1);
	      ReturnF(UDP_UNKNOWNSERVICE, n);
	    }
	  addr.netaddr.sin_port = sp->s_port;
	  /* Note: network routines seem to return values in net byte order */
	}
    }

  if (isdigit(*n))
    {
      addr.netaddr.sin_addr.s_addr = inet_addr(n);
      if (addr.netaddr.sin_addr.s_addr == -1)
	ReturnF(CONNECT_BADADDRESSFORMAT, n);
      free(n);

      /*
       * If they passed the address in dot notation, it's
       * unlikely they'll care to get the name later. If
       * they do, it'll be handled there.
       */
      addr.hostname = NULL;
    }
  else
    {
      char hnam[MAXHOSTNAMELEN];

      if (*n == '.')
	{
	  gethostname(hnam, MAXHOSTNAMELEN);
	  hp = gethostbyname(hnam);
	}
      else
	hp = gethostbyname(n);

      if (hp == NULL)
	ReturnF(CONNECT_UNKNOWNNAME, n);

      free(n);

      if (hp->h_addrtype != AF_INET)
	Return(CONNECT_DOMAINMISMATCH, name);

      addr.hostname = malloc(strlen(hp->h_name) + 1);
      if (addr.hostname != NULL)
	strcpy(addr.hostname, hp->h_name);

      memcpy(&addr.netaddr.sin_addr, hp->h_addr_list[0], 
	     sizeof(addr.netaddr.sin_addr));
    }

  /*
   * At this point, we have udpaddr filled up.
   * Now we look for an existing address which matches it.
   */
  s = inetlookup(&addr.netaddr);
  if (s != NULL)
    {
      if (addr.hostname != NULL)
	free(addr.hostname);
      *address = s->address;
      return OK;
    }

  if (!newaddress(&s, &addr))
    {
      *address = s->address;
      Return(CONNECT_NEWADDRESS, s->address);
    }

  if (addr.hostname != NULL)
    free(addr.hostname);
  return CHECK;
}

Trap udp_aton(address, name, length)
     Addr address;
     char *name;
     int length;
{
  udpaddr *a;
  struct hostent *hp;
  char *dot, port[10];

  if (address == wild)
    {
      strncat(name, "*", length - strlen(name) - 1);
      return OK;
    }

  a = addrlookup(address);
  if (a == NULL)
    Return(CONNECT_BADADDRESS, address);

  if (a->hostname == NULL)
    {
      if (nameResolve)
	hp = gethostbyaddr((char *) &(a->netaddr.sin_addr),
			   sizeof(a->netaddr.sin_addr), AF_INET);
      else
	hp = NULL;

      if (hp != NULL)
	{
	  strncat(name, hp->h_name, length - strlen(name) - 1);
	  a->hostname = malloc(strlen(hp->h_name) + 1);
	  if (a->hostname != NULL)
	    strcpy(a->hostname, hp->h_name);
	}
      else
	{
	  dot = inet_ntoa(a->netaddr.sin_addr);
	  strncat(name, dot, length - strlen(name) - 1);
	  a->hostname = malloc(strlen(dot) + 1);
	  if (a->hostname != NULL)
	    strcpy(a->hostname, dot);
	}
    }
  else
    strncat(name, a->hostname, length - strlen(name) - 1);

  sprintf(port, "/%d", ntohs(a->netaddr.sin_port));
  strncat(name, port, length - strlen(name) - 1);
  return OK;
}

/*
 * Since we are connectionless, the "to" address is irrelevant.
 *
 * ERRORS:
 *
 * CONNECT_ADDRESSOPENED: WARNING
 * The connection requested to be opened (%d) already was.
 *
 * UDP_SOCKET: FATAL
 * Socket creation failed with error %d.
 *
 * UDP_BIND: FATAL
 * The address could not be bound to a socket (error %d).
 */
Trap udp_open(from, to)
     Addr from, to;
{
  int s;
  udpaddr *a;

  a = addrlookup(from);
  if (a == NULL)
    Return(CONNECT_BADADDRESS, from);
  if (a->socket != 0)
    Return(CONNECT_ADDRESSOPENED, from);

  s = socket(AF_INET, SOCK_DGRAM, 0);
  if (s < 0)
    Return(UDP_SOCKET, errno);

  if (bind(s, (struct sockaddr *)&(a->netaddr), sizeof(a->netaddr)))
    {
      close(s);
      Return(UDP_BIND, errno); /* close won't disturb errno */
    }

  if (Connect_RegisterDescriptor(from, s))
    {
      close(s);
      return CHECK;
    }

  Connect_SelectInterest(s, Set, Clear, Clear);

  a->socket = s;
      
  return 0;
}

Trap udp_free(address)
     Addr address;
{
  return udp_close(address, wild);
}

Trap udp_close(from, to)
     Addr from, to;
{
  udpaddr *a;
  Addr which;

  if (to == wild)
    which = from;
  else
    which = to;

  a = addrlookup(which);
  if (a == NULL)
    Return(CONNECT_BADADDRESS, which);

  if (a->socket != 0)
    {
      Connect_FreeDescriptor(a->socket);
      close(a->socket);
      a->socket = 0;
    }

  freeaddr(a);
  Connect_KillAddress(which);
  return OK;
}

/*
 * ERRORS:
 *
 * CONNECT_ADDRESSNOTOPEN: FATAL
 * The operation on address %d failed because it is not open.
 *
 * CONNECT_BADDEST: FATAL
 * The destination address %d was insufficiently specified.
 *
 * UDP_SENDTO: FATAL
 * Sendto returned an unexpected error, %d.
 *
 * UDP_PARTIALSEND: WARNING
 * Sendto only sent part (%d bytes) of the packet.
 *
 * UDP_NOBROADCAST: FATAL
 * UDP broadcast is not supported.
 */
Trap udp_send(p)
     Packet *p;
{
  int cc;
  udpaddr *from, *to;

  if (p->Destination == wild)
    Return(UDP_NOBROADCAST, NULL);

  from = addrlookup(p->Source);
  if (from == NULL)
    Return(CONNECT_BADADDRESS, p->Source);
  if (from->socket == 0)
    Return(CONNECT_ADDRESSNOTOPEN, p->Source);

  to = addrlookup(p->Destination);
  if (to == NULL)
    Return(CONNECT_BADADDRESS, p->Destination);
  if (to->netaddr.sin_port == 0)
    Return(CONNECT_BADDEST, p->Destination);

  cc = sendto(from->socket, p->packet, p->length, 0,
	      (struct sockaddr *)&(to->netaddr), sizeof(to->netaddr));
  if (cc == -1)
    Return(UDP_SENDTO, errno);

  if (cc != p->length)
    Return(UDP_PARTIALSEND, cc);

  return OK;
}

Trap Udp_IP(addr, ip)
     Addr addr;
     u_long *ip;
{
  udpaddr *a;

  a = addrlookup(addr);
  if (a == NULL)
    Return(CONNECT_BADADDRESS, addr);

  *ip = a->netaddr.sin_addr.s_addr;
  return OK;
}

Trap Udp_Port(addr, port)
     Addr addr;
     u_short *port;
{
  udpaddr *a;

  a = addrlookup(addr);
  if (a == NULL)
    Return(CONNECT_BADADDRESS, addr);

  *port = ntohs(a->netaddr.sin_port);
  return OK;
}

/* Need to have a generic form of this to be called via Connect.
   Probably also for all the Udp_* routines. */
Trap Udp_NameResolve(val)
     int val;
{
  nameResolve = val;

  return OK;
}

/*
 * ERRORS:
 *
 * UDP_NREAD: FATAL
 * FIONREAD returned an unexpected error, %d.
 *
 * UDP_RECVFROM: FATAL
 * recvfrom returned an unexpected error, %d.
 *
 * UDP_DEADSOCKET: FATAL
 * The socket was expected to have data on it but did not.
 */
Trap udp_receive(s, a, p)
     int s;
     Addr a;
     Packet *p;
{
  int length, fromlen, cc;
  struct sockaddr_in from;
  char *buffer;
  udpaddr *source;
  udpaddr new;
  Trap trap = OK;

  /*
   * FIONREAD appears to return the number of bytes available to
   * be read. However, it may reflect the sum of bytes in multiple
   * packets. Read will only return one packet at a time, so
   * if there are multiple packets available all of FIONREAD bytes
   * will not be read. I don't know where this info is documented.
   * This is the behavior I have observed.
   */
  if (ioctl(s, FIONREAD, &length))
    Return(UDP_NREAD, errno);

  if (length != 0)
    {
      buffer = malloc(length);
      if (buffer == NULL)
	Return(GENERAL_MALLOC, length);

      fromlen = sizeof(from);
      cc = recvfrom(s, buffer, length, 0, (struct sockaddr *)&from, &fromlen);
      if (cc == -1)
	{
	  free(buffer);
	  Return(UDP_RECVFROM, errno);
	}

      source = inetlookup(&from);
      if (source == NULL)
	/* We haven't seen this address before. Make a record of
	   it. */
	{
	  memset(&new, 0, sizeof(new));
	  memcpy(&new.netaddr, &from, sizeof(new.netaddr));

	  if (newaddress(&source, &new))
	    {
	      /*
	       * XXX We could recover more gracefully here by having
	       * a reserved static address to which error replies
	       * could be directed, set to the address of the last
	       * packet dropped.
	       */
	      free(buffer);
	      return CHECK;
	    }

	  Error_Push(CONNECT_NEWADDRESS, source->address);
	  trap = CHECK;
	}

      p->Source = source->address;
      p->Destination = a;
      p->length = cc;
      p->packet = buffer;

      return trap;
    }

  /*
   * This shouldn't happen on an unconnected UDP socket.
   * So I don't really know what should be done here,
   * though maybe the socket should be closed.
   */
  Return(UDP_DEADSOCKET, a);
}
